// SPDX-FileCopyrightText: 2021-2024 Heiko Schaefer <heiko@schaefer.name>
// SPDX-FileCopyrightText: 2022 Nora Widdecke <mail@nora.pink>
// SPDX-FileCopyrightText: 2024 David Runge <dave@sleepmap.de>
// SPDX-License-Identifier: MIT OR Apache-2.0

use std::path::{Path, PathBuf};

use anyhow::{anyhow, Result};
use clap::Parser;
use sequoia_openpgp::packet::signature::SignatureBuilder;
use sequoia_openpgp::serialize::stream::{Armorer, LiteralWriter, Message, Signer};
use sequoia_openpgp::types::SignatureType;

use crate::util;

#[derive(Parser, Debug)]
pub struct SignCommand {
    #[arg(
        name = "card ident",
        short = 'c',
        long = "card",
        help = "Identifier of the card to use"
    )]
    pub ident: String,

    #[arg(
        name = "User PIN file",
        short = 'p',
        long = "user-pin",
        help = "Optionally, get User PIN from a file"
    )]
    pub user_pin: Option<PathBuf>,

    #[command(subcommand)]
    pub form: SignSubCommand,
}

#[derive(Debug, Parser, PartialEq)]
pub enum SignSubCommand {
    /// Create a detached signature
    Detached {
        /// Input file (stdin if unset)
        #[arg(name = "input")]
        input: Option<PathBuf>,

        /// Output file (stdout if unset)
        #[arg(name = "output", long = "output", short = 'o')]
        output: Option<PathBuf>,
    },

    /// Create a cleartext signature
    Cleartext {
        /// Input file (stdin if unset)
        #[arg(name = "input")]
        input: Option<PathBuf>,

        /// Output file (stdout if unset)
        #[arg(name = "output", long = "output", short = 'o')]
        output: Option<PathBuf>,
    },

    /// Create an inline signature
    Inline {
        /// Input file (stdin if unset)
        #[arg(name = "input")]
        input: Option<PathBuf>,

        /// Output file (stdout if unset)
        #[arg(name = "output", long = "output", short = 'o')]
        output: Option<PathBuf>,
    },
}

impl SignSubCommand {
    /// Get optional input and output Paths from the subcommand
    pub fn get_input_output(&self) -> (Option<&Path>, Option<&Path>) {
        match self {
            SignSubCommand::Detached { input, output }
            | SignSubCommand::Cleartext { input, output }
            | SignSubCommand::Inline { input, output } => (input.as_deref(), output.as_deref()),
        }
    }

    /// Evaluate whether the subcommand is used for creating a cleartext signature
    pub fn is_cleartext(&self) -> bool {
        matches!(self, SignSubCommand::Cleartext { .. })
    }

    /// Evaluate whether the subcommand is used for creating a detached signature
    pub fn is_detached(&self) -> bool {
        matches!(self, SignSubCommand::Detached { .. })
    }
}

pub fn sign(command: SignCommand) -> Result<(), Box<dyn std::error::Error>> {
    sign_message(&command.ident, command.user_pin, command.form)
}

pub fn sign_message(
    ident: &str,
    pin_file: Option<PathBuf>,
    form: SignSubCommand,
) -> Result<(), Box<dyn std::error::Error>> {
    let cleartext = form.is_cleartext();
    let detached = form.is_detached();
    let (input, output) = form.get_input_output();

    let mut input = util::open_or_stdin(input)?;

    let mut open = util::open_card(ident)?;
    let mut card = open.transaction()?;

    if card.fingerprints()?.signature().is_none() {
        return Err(anyhow!("Can't sign: this card has no key in the signing slot.").into());
    }

    let user_pin = util::get_pin(&mut card, pin_file, crate::ENTER_USER_PIN)?;

    let mut sign = util::verify_to_sign(&mut card, user_pin.as_deref())?;
    let s = sign.signer(&|| eprintln!("Touch confirmation needed for signing"))?;

    let sink = util::open_or_stdout(output)?;

    let mut message = Message::new(sink);
    if !cleartext {
        message = Armorer::new(message).build()?
    };

    let signer = if detached {
        Signer::new(message, s).detached()
    } else if cleartext {
        Signer::with_template(message, s, SignatureBuilder::new(SignatureType::Text)).cleartext()
    } else {
        Signer::new(message, s)
    }
    .build()?;

    let mut writer = if detached || cleartext {
        signer
    } else {
        LiteralWriter::new(signer).build()?
    };

    std::io::copy(&mut input, &mut writer)?;
    writer.finalize()?;

    Ok(())
}
